import typing
from abc import abstractmethod
from typing import TYPE_CHECKING
from dataclasses import dataclass
from collections.abc import Hashable

from beartype.door import TupleVariableTypeHint  # type: ignore
from beartype.door import TypeHint, ClassTypeHint, LiteralTypeHint, TypeVarTypeHint

from kirin.serialization.base.serializationunit import SerializationUnit

if TYPE_CHECKING:
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer

from typing_extensions import Never

from kirin.print import Printer
from kirin.lattice import (
    UnionMeta,
    SingletonMeta,
    BoundedLattice,
    IsSubsetEqMixin,
    SimpleMeetMixin,
)

from .abc import Attribute, LatticeAttributeMeta
from ._types import _TypeAttribute


class TypeAttributeMeta(LatticeAttributeMeta):
    """Metaclass for type attributes."""

    pass


class SingletonTypeMeta(TypeAttributeMeta, SingletonMeta):
    """Metaclass for singleton type attributes.

    Singleton type attributes are attributes that have only one instance.

    Examples:
    - `AnyType`
    - `BottomType`
    """

    pass


class UnionTypeMeta(TypeAttributeMeta, UnionMeta):
    pass


@dataclass
class TypeAttribute(
    _TypeAttribute,
    SimpleMeetMixin["TypeAttribute"],
    IsSubsetEqMixin["TypeAttribute"],
    BoundedLattice["TypeAttribute"],
    metaclass=TypeAttributeMeta,
):

    @classmethod
    def top(cls) -> "TypeAttribute":
        return AnyType()

    @classmethod
    def bottom(cls) -> "TypeAttribute":
        return BottomType()

    def join(self, other: "TypeAttribute") -> "TypeAttribute":
        if self.is_subseteq(other):
            return other
        elif other.is_subseteq(self):
            return self
        elif isinstance(other, TypeAttribute):
            return Union(self, other)
        return AnyType()  # don't know how to join

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self, prefix="!")

    def __or__(self, other: "TypeAttribute"):
        return self.join(other)

    def __eq__(self, value: object) -> bool:
        return isinstance(value, TypeAttribute) and self.is_structurally_equal(value)

    @abstractmethod
    def is_structurally_equal(
        self, other: "Attribute", context: dict | None = None
    ) -> bool: ...

    @abstractmethod
    def __hash__(self) -> int: ...


@typing.final
@dataclass(eq=False)
class AnyType(TypeAttribute, metaclass=SingletonTypeMeta):
    name = "Any"

    def is_subseteq_TypeVar(self, other: "TypeVar") -> bool:
        return self.is_subseteq(other.bound)

    def __hash__(self) -> int:
        return id(self)

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        return isinstance(other, AnyType)

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_anytype(self)

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "AnyType":
        return deserializer.deserialize_anytype(serUnit)


@typing.final
@dataclass(eq=False)
class BottomType(TypeAttribute, metaclass=SingletonTypeMeta):
    name = "Bottom"

    def is_subseteq(self, other: TypeAttribute) -> bool:
        if isinstance(other, TypeVar):
            return self.is_subseteq(other.bound)
        return True

    def __hash__(self) -> int:
        return id(self)

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        return isinstance(other, BottomType)

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_bottomtype(self)

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "BottomType":
        return deserializer.deserialize_bottomtype(serUnit)


class PyClassMeta(TypeAttributeMeta):

    def __init__(self, *args, **kwargs):
        super(PyClassMeta, self).__init__(*args, **kwargs)
        self._cache = {}

    def __call__(self, typ, *, display_name: str | None = None, prefix="py"):
        display_name = display_name if display_name is not None else typ.__name__

        if typ is typing.Any:
            return AnyType()
        elif typ is typing.NoReturn or typ is Never:
            return BottomType()
        elif typ is typing.Tuple:
            typ = tuple
        elif typ is typing.List:
            typ = list
        elif isinstance(typ, TypeVar):
            return hint2type(typ)

        if isinstance(typ, type) and typ in self._cache:
            obj = self._cache[typ]
            if display_name != obj.display_name or prefix != obj.prefix:
                raise ValueError(
                    f"Type {typ} already registered to {obj.prefix}.{obj.display_name}"
                )

            return self._cache[typ]

        instance = super(PyClassMeta, self).__call__(
            typ, display_name=display_name, prefix=prefix
        )
        self._cache[typ] = instance
        return instance


PyClassType = typing.TypeVar("PyClassType")


@typing.final
@dataclass(eq=False)
class PyClass(TypeAttribute, typing.Generic[PyClassType], metaclass=PyClassMeta):
    name = "PyClass"
    typ: type[PyClassType]
    display_name: str
    prefix: str

    def __init__(
        self,
        typ: type[PyClassType],
        *,
        display_name: str | None = None,
        prefix: str = "py",
    ) -> None:
        self.typ = typ
        self.display_name = display_name if display_name is not None else typ.__name__
        self.prefix = prefix

    def is_subseteq_PyClass(self, other: "PyClass") -> bool:
        return issubclass(self.typ, other.typ)

    def is_subseteq_Union(self, other: "Union") -> bool:
        return any(self.is_subseteq(t) for t in other.types)

    def is_subseteq_Generic(self, other: "Generic") -> bool:
        # NOTE: subclass without generics is just generic with all any parameters
        Any = AnyType()
        return (
            self.is_subseteq(other.body)
            and all(Any.is_subseteq(bound) for bound in other.vars)
            and (other.vararg is None or Any.is_subseteq(other.vararg.typ))
        )

    def is_subseteq_TypeVar(self, other: "TypeVar") -> bool:
        return self.is_subseteq(other.bound)

    def __hash__(self) -> int:
        return hash((PyClass, self.typ))

    def __repr__(self) -> str:
        return self.typ.__name__

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print(f"!{self.prefix}.", self.display_name)

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        return isinstance(other, PyClass) and self.typ == other.typ

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_pyclass(self)

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "PyClass":
        return deserializer.deserialize_pyclass(serUnit)


class LiteralMeta(TypeAttributeMeta):

    def __init__(self, *args, **kwargs):
        super(LiteralMeta, self).__init__(*args, **kwargs)
        self._cache = {}

    def __call__(self, data, datatype=None):
        if isinstance(data, TypeAttribute):
            return data  # already a type
        elif not isinstance(data, Hashable):
            raise ValueError("Literal data must be hashable")
        elif (data, datatype) in self._cache:
            return self._cache[(data, datatype)]

        instance = super(LiteralMeta, self).__call__(data, datatype)
        self._cache[(data, datatype)] = instance
        return instance


LiteralType = typing.TypeVar("LiteralType")


@typing.final
@dataclass(eq=False)
class Literal(TypeAttribute, typing.Generic[LiteralType], metaclass=LiteralMeta):
    name = "Literal"
    data: LiteralType
    type: TypeAttribute

    """type of the literal, this is useful when the Python type of
    data does not represent the type in IR, e.g Literal(1, types.Int32)
    """

    def __init__(self, data: LiteralType, datatype: TypeAttribute | None = None):
        self.data = data
        self.type = datatype or PyClass(type(data))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Literal):
            return False
        return self.data == other.data and self.type == other.type

    def is_subseteq_TypeVar(self, other: "TypeVar") -> bool:
        return self.is_subseteq(other.bound)

    def is_subseteq_Union(self, other: "Union") -> bool:
        return any(self.is_subseteq(t) for t in other.types)

    def is_subseteq_Literal(self, other: "Literal") -> bool:
        return self.data == other.data and self.type.is_subseteq(other.type)

    def is_subseteq_fallback(self, other: TypeAttribute) -> bool:
        return self.type.is_subseteq(other)

    def __hash__(self) -> int:
        return hash((Literal, self.data))

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print("Literal(", repr(self.data), ",", self.type, ")")

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        return (
            isinstance(other, Literal)
            and self.data == other.data
            and self.type.is_structurally_equal(other.type, context=context)
        )

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_literal(self)

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "Literal":
        return deserializer.deserialize_literal(serUnit)


@typing.final
@dataclass(eq=False)
class Union(TypeAttribute, metaclass=UnionTypeMeta):
    name = "Union"
    types: frozenset[TypeAttribute]

    def __init__(
        self,
        typ_or_set: TypeAttribute | typing.Iterable[TypeAttribute],
        *typs: TypeAttribute,
    ):
        if isinstance(typ_or_set, TypeAttribute):
            params: typing.Iterable[TypeAttribute] = (typ_or_set, *typs)
        else:
            params = typ_or_set
            assert not typs, "Cannot pass multiple arguments when passing a set"

        types: frozenset[TypeAttribute] = frozenset()
        for typ in params:
            if isinstance(typ, Union):
                types = types.union(typ.types)
            else:
                types = types.union({typ})
        self.types = types

    def is_subseteq_fallback(self, other: TypeAttribute) -> bool:
        return all(t.is_subseteq(other) for t in self.types)

    def join(self, other: TypeAttribute) -> TypeAttribute:
        if self.is_subseteq(other):
            return other
        elif other.is_subseteq(self):
            return self
        elif isinstance(other, Union):
            return Union(self.types | other.types)
        elif isinstance(other, TypeAttribute):
            return Union(self.types | {other})
        return BottomType()

    def meet(self, other: TypeAttribute) -> TypeAttribute:
        if self.is_subseteq(other):
            return self
        elif other.is_subseteq(self):
            return other
        elif isinstance(other, Union):
            return Union(self.types & other.types)
        elif isinstance(other, TypeAttribute):
            return Union(self.types & {other})
        return BottomType()

    def __hash__(self) -> int:
        return hash((Union, self.types))

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self, prefix="!")
        printer.print_seq(self.types, delim=", ", prefix="[", suffix="]")

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        return isinstance(other, Union) and self.types == other.types

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_union(self)

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "Union":
        return deserializer.deserialize_union(serUnit)


@typing.final
@dataclass(eq=False)
class TypeVar(TypeAttribute):
    name = "TypeVar"
    varname: str
    bound: TypeAttribute

    def __init__(self, name: str, bound: TypeAttribute | None = None):
        self.varname = name
        self.bound = bound or AnyType()

    def is_subseteq_TypeVar(self, other: "TypeVar") -> bool:
        return self.bound.is_subseteq(other.bound)

    def is_subseteq_Union(self, other: Union) -> bool:
        return any(self.is_subseteq(t) for t in other.types)

    def is_subseteq_fallback(self, other: TypeAttribute) -> bool:
        return self.bound.is_subseteq(other)

    def __hash__(self) -> int:
        return hash((TypeVar, self.varname, self.bound))

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print(f"~{self.varname}")
        if self.bound is not self.bound.top():
            printer.plain_print(" : ")
            printer.print(self.bound)

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        return (
            isinstance(other, TypeVar)
            and self.varname == other.varname
            and self.bound.is_structurally_equal(other.bound, context=context)
        )

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_typevar(self)

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "TypeVar":
        return deserializer.deserialize_typevar(serUnit)


@typing.final
@dataclass
class Vararg(Attribute):
    name = "Vararg"
    typ: TypeAttribute

    def __hash__(self) -> int:
        return hash((Vararg, self.typ))

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Vararg):
            return False

        return self.typ == value.typ

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print("*")
        printer.print(self.typ)

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        return isinstance(other, Vararg) and self.typ.is_structurally_equal(other.typ)

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_vararg(self)

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "Vararg":
        return deserializer.deserialize_vararg(serUnit)


TypeVarValue: typing.TypeAlias = TypeAttribute | Vararg | list
TypeOrVararg: typing.TypeAlias = TypeAttribute | Vararg


@typing.final
@dataclass(eq=False)
class Generic(TypeAttribute, typing.Generic[PyClassType]):
    name = "Generic"
    body: PyClass[PyClassType]
    vars: tuple[TypeAttribute, ...]
    vararg: Vararg | None = None

    def __init__(
        self,
        body: type[PyClassType] | PyClass[PyClassType],
        *vars: TypeAttribute | list | Vararg,
    ):
        if isinstance(body, PyClass):
            self.body = body
        else:
            self.body = PyClass(body)
        self.vars, self.vararg = _split_type_args(vars)

    def is_subseteq_Literal(self, other: Literal) -> bool:
        return False

    def is_subseteq_PyClass(self, other: PyClass) -> bool:
        return self.body.is_subseteq(other)

    def is_subseteq_Union(self, other: Union) -> bool:
        return any(self.is_subseteq(t) for t in other.types)

    def is_subseteq_TypeVar(self, other: TypeVar) -> bool:
        return self.body.is_subseteq(other.bound)

    def is_subseteq_Generic(self, other: "Generic") -> bool:
        if other.vararg is None:
            return (
                self.body.is_subseteq(other.body)
                and len(self.vars) == len(other.vars)
                and all(v.is_subseteq(o) for v, o in zip(self.vars, other.vars))
            )
        else:
            return (
                self.body.is_subseteq(other.body)
                and len(self.vars) >= len(other.vars)
                and all(v.is_subseteq(o) for v, o in zip(self.vars, other.vars))
                and all(
                    v.is_subseteq(other.vararg.typ)
                    for v in self.vars[len(other.vars) :]
                )
                and (
                    self.vararg is None or self.vararg.typ.is_subseteq(other.vararg.typ)
                )
            )

    def __hash__(self) -> int:
        return hash((Generic, self.body, self.vars, self.vararg))

    def __repr__(self) -> str:
        if self.vararg is None:
            return f"{self.body}[{', '.join(map(repr, self.vars))}]"
        else:
            return f"{self.body}[{', '.join(map(repr, self.vars))}, {self.vararg}, ...]"

    def print_impl(self, printer: Printer) -> None:
        printer.print(self.body)
        printer.plain_print("[")
        if self.vars:
            printer.print_seq(self.vars)
        if self.vararg is not None:
            if self.vars:
                printer.plain_print(", ")
            printer.print(self.vararg.typ)
            printer.plain_print(", ...")
        printer.plain_print("]")

    def __getitem__(self, typ: TypeVarValue | tuple[TypeVarValue, ...]) -> "Generic":
        return self.where(typ)

    def where(self, typ: TypeVarValue | tuple[TypeVarValue, ...]) -> "Generic":
        if isinstance(typ, tuple):
            typs = typ
        else:
            typs = (typ,)

        args, vararg = _split_type_args(typs)
        if self.vararg is None and vararg is None:
            assert len(args) <= len(
                self.vars
            ), "Number of type arguments does not match"
            if all(v.is_subseteq(bound) for v, bound in zip(args, self.vars)):
                return Generic(self.body, *args, *self.vars[len(args) :])
            else:
                raise TypeError("Type arguments do not match")
        elif self.vararg is not None and vararg is None:
            assert len(args) >= len(
                self.vars
            ), "Number of type arguments does not match"
            if all(v.is_subseteq(bound) for v, bound in zip(args, self.vars)) and all(
                v.is_subseteq(self.vararg.typ) for v in args[len(self.vars) :]
            ):
                return Generic(self.body, *args)
        elif self.vararg is not None and vararg is not None:
            if len(args) < len(self.vars):
                if (
                    all(v.is_subseteq(bound) for v, bound in zip(args, self.vars))
                    and all(
                        vararg.typ.is_subseteq(bound)
                        for bound in self.vars[len(args) :]
                    )
                    and vararg.typ.is_subseteq(self.vararg.typ)
                ):
                    return Generic(self.body, *args, vararg)
            else:
                if (
                    all(v.is_subseteq(bound) for v, bound in zip(args, self.vars))
                    and all(v.is_subseteq(vararg.typ) for v in args[len(self.vars) :])
                    and vararg.typ.is_subseteq(self.vararg.typ)
                ):
                    return Generic(self.body, *args, vararg)
        raise TypeError("Type arguments do not match")

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        if not isinstance(other, Generic):
            return False
        if self.body != other.body:
            return False
        if len(self.vars) != len(other.vars):
            return False
        if any(
            not v.is_structurally_equal(o, context=context)
            for v, o in zip(self.vars, other.vars)
        ):
            return False
        if self.vararg is None and other.vararg is None:
            return True
        if self.vararg is not None and other.vararg is not None:
            return self.vararg.typ.is_structurally_equal(
                other.vararg.typ, context=context
            )
        return False

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_generic(self)

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "Generic":
        return deserializer.deserialize_generic(serUnit)


def _typeparams_list2tuple(args: tuple[TypeVarValue, ...]) -> tuple[TypeOrVararg, ...]:
    "provides the syntax sugar [A, B, C] type Generic(tuple, A, B, C)"
    return tuple(Generic(tuple, *arg) if isinstance(arg, list) else arg for arg in args)


def _split_type_args(
    args: tuple[TypeVarValue, ...],
) -> tuple[tuple[TypeAttribute, ...], Vararg | None]:
    args = _typeparams_list2tuple(args)
    if args is None or len(args) == 0:
        return (), None

    if isinstance(args[-1], Vararg):
        xs = args[:-1]
        if is_tuple_of(xs, TypeAttribute):
            return xs, args[-1]
        else:
            raise TypeError("Multiple varargs are not allowed")
    elif is_tuple_of(args, TypeAttribute):
        return args, None
    raise TypeError("Vararg must be the last argument")


T = typing.TypeVar("T")


def is_tuple_of(xs: tuple, typ: type[T]) -> typing.TypeGuard[tuple[T, ...]]:
    return all(isinstance(x, typ) for x in xs)


def hint2type(hint) -> TypeAttribute:
    if isinstance(hint, TypeAttribute):
        return hint
    elif hint is None:
        return PyClass(type(None))

    bear_hint = TypeHint(hint)
    if isinstance(bear_hint, LiteralTypeHint):
        return Literal(typing.get_args(hint)[0])
    elif isinstance(bear_hint, TypeVarTypeHint):
        return TypeVar(
            hint.__name__,
            hint2type(hint.__bound__) if hint.__bound__ else None,
        )
    elif isinstance(bear_hint, ClassTypeHint):
        return PyClass(hint)
    elif isinstance(bear_hint, TupleVariableTypeHint):
        if len(bear_hint.args) != 1:
            raise TypeError("Tuple hint must have exactly one argument")
        return Generic(tuple, Vararg(hint2type(bear_hint.args[0])))

    origin: type | None = typing.get_origin(hint)
    if origin is None:  # non-generic
        return PyClass(hint)

    body = PyClass(origin)
    args = typing.get_args(hint)
    params = []
    for arg in args:
        if isinstance(arg, typing.Sequence):
            params.append([hint2type(elem) for elem in arg])
        else:
            params.append(hint2type(arg))
    return Generic(body, *params)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Iterable
from dataclasses import dataclass

if TYPE_CHECKING:
    import ast

    from kirin import lowering
    from kirin.ir.group import DialectGroup
    from kirin.lowering import FromPythonAST
    from kirin.interp.table import Signature, BoundedDef
    from kirin.lowering.python.dialect import LoweringTransform


@dataclass
class CallLoweringFunction:
    parent: FromPythonAST
    func: LoweringTransform

    def __call__(
        self, state: "lowering.State[ast.AST]", node: "ast.Call"
    ) -> "lowering.Result":
        return self.func(self.parent, state, node)


@dataclass
class LoweringRegistry:
    ast_table: dict[str, FromPythonAST]
    callee_table: dict[object, CallLoweringFunction]


@dataclass
class Registry:
    """Proxy class to build different registries from a dialect group."""

    dialects: "DialectGroup"
    """The dialect group to build the registry from."""

    def ast(self, keys: Iterable[str]) -> LoweringRegistry:
        """select the dialect lowering interpreters for the given key.

        Args:
            keys (Iterable[str]): the keys to search for in the dialects

        Returns:
            a map of dialects to their lowering interpreters
        """
        ret: dict[str, FromPythonAST] = {}
        callee_table: dict[object, CallLoweringFunction] = {}
        from_ast = None
        for dialect in self.dialects.data:
            for key in keys:
                if key in dialect.lowering:
                    from_ast = dialect.lowering[key]
                    break

            if from_ast is None:
                msg = ",".join(keys)
                raise KeyError(f"Lowering not found for {msg}")

            for name in from_ast.names:
                if name.startswith("lower_Call_"):
                    continue  # ignore the Call nodes
                if name in ret:
                    raise KeyError(f"Lowering {name} already exists")
                ret[name] = from_ast

            for obj, trans in from_ast.callee_table.items():
                if obj in callee_table:
                    raise KeyError(f"Lowering {obj} already exists")
                callee_table[obj] = CallLoweringFunction(from_ast, trans.func)
        return LoweringRegistry(ret, callee_table)

    def interpreter(self, keys: Iterable[str]) -> dict["Signature", "BoundedDef"]:
        """select the dialect interpreter for the given key.

        Args:
            keys (Iterable[str]): the keys to search for in the dialects

        Returns:
            a map of statement signatures to their interpretation functions,
            and a map of dialects to their fallback interpreters.
        """
        from kirin.interp.table import BoundedDef

        registry: dict[Signature, BoundedDef] = {}
        for dialect in self.dialects.data:
            dialect_table = None
            for key in keys:
                if key not in dialect.interps:
                    continue

                dialect_table = dialect.interps[key]
                for _, member in inspect.getmembers(dialect_table):
                    if isinstance(member, BoundedDef):
                        for sig in member.signature:
                            registry[sig] = member
                break
        return registry



from kirin import ir, emit, interp

from .stmts import For, Yield, IfElse
from ._dialect import dialect


@dialect.register(key="emit.julia")
class JuliaEmit(interp.MethodTable):
    @interp.impl(IfElse)
    def ifelse_(self, emit_: emit.Julia, frame: emit.JuliaFrame, node: IfElse):
        then_block = node.then_body.blocks[0]
        else_block = node.else_body.blocks[0]

        cond = frame.get(node.cond)
        for result in node.results:  # declare the result variables
            frame.write_line(f"local {frame.ssa[result]}")

        frame.write_line(f"if {frame.get(node.cond)}")
        with frame.indent():
            frame.set(then_block.args[0], frame.ssa[then_block.args[0]])
            frame.write_line(f"{frame.ssa[then_block.args[0]]} = {cond}")
            self.walk_yield_block(emit_, frame, node, then_block)

        frame.write_line("else")
        with frame.indent():
            frame.set(else_block.args[0], frame.ssa[else_block.args[0]])
            frame.write_line(f"{frame.ssa[else_block.args[0]]} = {cond}")
            self.walk_yield_block(emit_, frame, node, else_block)
        frame.write_line("end")
        return tuple(frame.ssa[result] for result in node.results)

    @interp.impl(For)
    def for_(self, emit: emit.Julia, frame: emit.JuliaFrame, node: For):
        block = node.body.blocks[0]
        index = block.args[0]
        frame.current_block = block

        for result in node.results:  # declare the result variables
            frame.write_line(f"local {frame.ssa[result]}")

        frame.write_line(f"for {frame.ssa[index]} in {frame.get(node.iterable)}")
        with frame.indent():
            for arg, value in zip(block.args[1:], node.initializers):
                frame.write_line(f"{frame.ssa[arg]} = {frame.get(value)}")
            for arg in block.args:
                frame.set(arg, frame.ssa[arg])
            self.walk_yield_block(emit, frame, node, block)
        frame.write_line("end")

        return tuple(frame.ssa[result] for result in node.results)

    @classmethod
    def walk_yield_block(
        cls,
        emit_: emit.Julia,
        frame: emit.JuliaFrame,
        node: ir.Statement,
        block: ir.Block,
    ):
        for stmt in block.stmts:
            frame.current_stmt = stmt
            if isinstance(stmt, Yield):
                for result, value in zip(node.results, stmt.values):
                    frame.write_line(f"{frame.ssa[result]} = {frame.get(value)}")
                continue

            stmt_results = emit_.frame_eval(frame, stmt)
            if isinstance(stmt_results, tuple):
                frame.set_values(stmt._results, stmt_results)
            elif stmt_results is None:
                continue
            else:
                raise interp.InterpreterError(
                    "unexpected statement result, expected tuple or None"
                )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar, TypeAlias
from dataclasses import dataclass

from kirin.ir import Region, SSAValue, Statement, DialectGroup

from .exception import BuildError

if TYPE_CHECKING:
    from .state import State


Result: TypeAlias = SSAValue | tuple[SSAValue, ...] | Statement | None
"""Result of lowering a node.
This is used to indicate that the node can be lowered to a SSAValue or None.

If the node is corresponding to a single statement that has a single result value,
the result can also be a Statement for convenience.

If the node can be assigned to a variable syntax-wise, it returns the SSAValue.
If the node cannot be assigned to a variable, it returns None.
"""

EntryNodeType = TypeVar("EntryNodeType")
ASTNodeType = TypeVar("ASTNodeType")


@dataclass
class LoweringABC(ABC, Generic[ASTNodeType]):
    """Base class for lowering.

    This class is used to lower the AST nodes to IR.
    It contains the lowering process and the state of the lowering process.
    """

    dialects: DialectGroup
    """dialects to lower to"""

    @abstractmethod
    def run(
        self,
        stmt: ASTNodeType,
        *,
        source: str | None = None,
        globals: dict[str, Any] | None = None,
        file: str | None = None,
        lineno_offset: int = 0,
        col_offset: int = 0,
        compactify: bool = True,
    ) -> Region: ...

    @abstractmethod
    def visit(self, state: State[ASTNodeType], node: ASTNodeType) -> Result:
        """Entry point of AST visitors.

        Args:
            state: lowering state
            node: AST node to be lowered
        Returns:
            SSAValue: if the node can be assigned to a variable syntax-wise,
                what is the `SSAValue`.
            Statement: if the node is a single statement that has a single result value.
                This is equivalent to returning `stmt.results[0]`.
            None: If the node cannot be assigned to a variable syntax-wise.
        Raises:
            lowering.BuildError: if the node cannot be lowered.
        """
        ...

    @abstractmethod
    def lower_literal(self, state: State[ASTNodeType], value) -> SSAValue: ...

    @dataclass
    class Result:
        data: Any

        ExpectT = TypeVar("ExpectT")

        def expect(self, typ: type[ExpectT]) -> ExpectT:
            if not isinstance(self.data, typ):
                raise BuildError(f"expected {typ}, got {type(self.data)}")
            return self.data

    @abstractmethod
    def lower_global(
        self, state: State[ASTNodeType], node: ASTNodeType
    ) -> LoweringABC.Result:
        """Transform a given global expression to a SSAValue.

        This method is overridden by the subclass to transform a given global
        AST expression to a value as `LoweringABC.Result`.

        The subclass must implement this method to transform a given global
        AST expression to a SSAValue.
        """
        ...

    def lower_global_no_raise(
        self, state: State[ASTNodeType], node: ASTNodeType
    ) -> LoweringABC.Result | None:
        """Transform a given global expression to a SSAValue.

        This method can be overridden by the subclass to transform a given global
        AST expression to a SSAValue.
        """
        try:
            return self.lower_global(state, node)
        except BuildError:
            return None



import inspect

from kirin.prelude import basic_no_opt


def test_docstring():
    @basic_no_opt
    def some_func(x):
        "Some kernel function"
        return x + 1

    assert inspect.getdoc(some_func) == "Some kernel function"
    assert some_func(1) == 2



import io
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Union,
    Literal,
    TypeVar,
    Callable,
    Iterable,
    Generator,
)
from contextlib import contextmanager
from dataclasses import field, dataclass

from rich.theme import Theme
from rich.console import Console

from kirin.idtable import IdTable
from kirin.print.printable import Printable

if TYPE_CHECKING:
    from kirin import ir


DEFAULT_THEME = {
    "dark": Theme(
        {
            "dialect": "dark_blue",
            "type": "dark_blue",
            "comment": "bright_black",
            "keyword": "red",
            "symbol": "cyan",
            "warning": "yellow",
            "string": "green",
            "irrational": "default",
            "number": "default",
        }
    ),
    "light": Theme(
        {
            "dialect": "blue",
            "type": "blue",
            "comment": "bright_black",
            "keyword": "red",
            "symbol": "cyan",
            "warning": "yellow",
            "string": "green",
            "irrational": "magenta",
        }
    ),
}


@dataclass
class PrintState:
    ssa_id: IdTable["ir.SSAValue"] = field(default_factory=IdTable["ir.SSAValue"])
    block_id: IdTable["ir.Block"] = field(
        default_factory=lambda: IdTable["ir.Block"](prefix="^")
    )
    indent: int = 0
    result_width: int = 0
    indent_marks: list[int] = field(default_factory=list)
    result_width: int = 0
    "SSA-value column width in printing"
    rich_style: str | None = None
    rich_highlight: bool | None = False
    messages: list[str] = field(default_factory=list)


IOType = TypeVar("IOType", bound=IO)


def _default_console():
    return Console(force_jupyter=False)


@dataclass
class Printer:
    """A IR pretty printer build on top of Rich console."""

    console: Console = field(default_factory=_default_console)
    """Rich console"""
    analysis: dict["ir.SSAValue", Any] | None = None
    """Analysis results"""
    hint: str | None = None
    """key of the SSAValue hint to print"""
    state: PrintState = field(default_factory=PrintState)
    """Printing state"""
    show_indent_mark: bool = field(default=True, kw_only=True)
    "Whether to show indent marks, e.g │"
    theme: Theme | dict | Literal["dark", "light"] = field(default="dark", kw_only=True)
    "Theme to use for printing"

    def __post_init__(self):
        if isinstance(self.theme, dict):
            self.theme = Theme(self.theme)
        elif isinstance(self.theme, str):
            self.theme = DEFAULT_THEME[self.theme]
        self.console.push_theme(self.theme)

    def print(self, object):
        """entry point for printing an object

        Args:
            object: object to print.
        """
        if isinstance(object, Printable):
            object.print_impl(self)
        else:
            fn = getattr(self, f"print_{object.__class__.__name__}", None)
            if fn is None:
                raise NotImplementedError(
                    f"Printer for {object.__class__.__name__} not found"
                )
            fn(object)

    def print_name(
        self, node: Union["ir.Attribute", "ir.Statement"], prefix: str = ""
    ) -> None:
        """print the name of a node

        Args:
            node(ir.Attribute | ir.Statement): node to print
            prefix(str): prefix to print before the name, default to ""
        """
        self.print_dialect_path(node, prefix=prefix)
        if node.dialect:
            self.plain_print(".")
        self.plain_print(node.name)

    def print_stmt(self, node: "ir.Statement"):
        if node._results:
            result_str = self.result_str(node._results)
            self.plain_print(result_str.rjust(self.state.result_width), " = ")
        elif self.state.result_width:
            self.plain_print(" " * self.state.result_width, "   ")
        with self.indent(self.state.result_width + 3, mark=True):
            self.print(node)
            with self.rich(style="warning"):
                self.print_analysis(*node._results, prefix=" # ---> ")
            with self.rich(style="comment"):
                self.print_hint(*node._results)

    def print_hint(
        self,
        *values: "ir.SSAValue",
        prefix: str = " // hint<",
        suffix: str = ">",
    ):
        if not self.hint or not values:
            return

        self.plain_print(prefix)
        self.plain_print(self.hint)
        for idx, item in enumerate(values):
            if idx > 0:
                self.plain_print(", ")

            self.plain_print("=")
            if item.hints.get(self.hint) is not None:
                self.plain_print(repr(item.hints.get(self.hint)))
            else:
                self.plain_print("missing")
        self.plain_print(suffix)

    def print_analysis(
        self,
        *values: "ir.SSAValue",
        prefix: str = "",
    ):
        if self.analysis is None or not values:
            return

        self.plain_print(prefix)
        for idx, value in enumerate(values):
            if idx > 0:
                self.plain_print(", ")

            if result := self.analysis.get(value):
                self.plain_print(repr(result))
            else:
                self.plain_print("missing")

    def print_dialect_path(
        self, node: Union["ir.Attribute", "ir.Statement"], prefix: str = ""
    ) -> None:
        """print the dialect path of a node.

        Args:
            node(ir.Attribute | ir.Statement): node to print
            prefix(str): prefix to print before the dialect path, default to ""
        """
        if node.dialect:  # not None
            self.plain_print(prefix)
            self.plain_print(node.dialect.name, style="dialect")
        else:
            self.plain_print(prefix)

    def print_newline(self):
        """print a newline character.

        This method also prints any messages in the state for debugging.
        """
        self.plain_print("\n")

        if self.state.messages:
            for message in self.state.messages:
                self.plain_print(message)
                self.plain_print("\n")
            self.state.messages.clear()
        self.print_indent()

    def print_indent(self):
        """print the current indentation level optionally with indent marks."""
        indent_str = ""
        if self.show_indent_mark and self.state.indent_marks:
            indent_str = "".join(
                "│" if i in self.state.indent_marks else " "
                for i in range(self.state.indent)
            )
            with self.rich(style="comment"):
                self.plain_print(indent_str)
        else:
            indent_str = " " * self.state.indent
            self.plain_print(indent_str)

    def plain_print(self, *objects, sep="", end="", style=None, highlight=None):
        """print objects without any formatting.

        Args:
            *objects: objects to print

        Keyword Args:
            sep(str): separator between objects, default to ""
            end(str): end character, default to ""
            style(str): style to use, default to None
            highlight(bool): whether to highlight the text, default to None
        """
        self.console.out(
            *objects,
            sep=sep,
            end=end,
            style=style or self.state.rich_style,
            highlight=highlight or self.state.rich_highlight,
        )

    ElemType = TypeVar("ElemType")

    def print_seq(
        self,
        seq: Iterable[ElemType],
        *,
        emit: Callable[[ElemType], None] | None = None,
        delim: str = ", ",
        prefix: str = "",
        suffix: str = "",
        style=None,
        highlight=None,
    ) -> None:
        """print a sequence of objects.

        Args:
            seq(Iterable[ElemType]): sequence of objects to print

        Keyword Args:
            emit(Callable[[ElemType], None]): function to print each element, default to None
            delim(str): delimiter between elements, default to ", "
            prefix(str): prefix to print before the sequence, default to ""
            suffix(str): suffix to print after the sequence, default to ""
            style(str): style to use, default to None
            highlight(bool): whether to highlight the text, default to None
        """
        emit = emit or self.print
        self.plain_print(prefix, style=style, highlight=highlight)
        for idx, item in enumerate(seq):
            if idx > 0:
                self.plain_print(delim, style=style)
            emit(item)
        self.plain_print(suffix, style=style, highlight=highlight)

    KeyType = TypeVar("KeyType")
    ValueType = TypeVar("ValueType", bound=Printable)

    def print_mapping(
        self,
        elems: dict[KeyType, ValueType],
        *,
        emit: Callable[[ValueType], None] | None = None,
        delim: str = ", ",
    ) -> None:
        """print a mapping of key-value pairs.

        Args:
            elems(dict[KeyType, ValueType]): mapping to print

        Keyword Args:
            emit(Callable[[ValueType], None]): function to print each value, default to None
            delim(str): delimiter between key-value pairs, default to ", "
        """
        emit = emit or self.print
        for i, (key, value) in enumerate(elems.items()):
            if i > 0:
                self.plain_print(delim)
            self.plain_print(f"{key}=")
            emit(value)

    def result_width(self, stmts: Iterable["ir.Statement"]) -> int:
        """return the maximum width of the result column for a sequence of statements.

        Args:
            stmts(Iterable[ir.Statement]): sequence of statements

        Returns:
            int: maximum width of the result column
        """
        result_width = 0
        for stmt in stmts:
            result_width = max(result_width, len(self.result_str(stmt._results)))
        return result_width

    @contextmanager
    def align(self, width: int) -> Generator[PrintState, Any, None]:
        """align the result column width, and restore it after the context.

        Args:
            width(int): width of the column

        Yields:
            PrintState: the state with the new column width
        """
        old_width = self.state.result_width
        self.state.result_width = width
        try:
            yield self.state
        finally:
            self.state.result_width = old_width

    @contextmanager
    def indent(
        self, increase: int = 2, mark: bool | None = None
    ) -> Generator[PrintState, Any, None]:
        """increase the indentation level, and restore it after the context.

        Args:
            increase(int): amount to increase the indentation level, default to 2
            mark(bool): whether to mark the indentation level, default to None

        Yields:
            PrintState: the state with the new indentation level.
        """
        mark = mark if mark is not None else self.show_indent_mark
        self.state.indent += increase
        if mark:
            self.state.indent_marks.append(self.state.indent)
        try:
            yield self.state
        finally:
            self.state.indent -= increase
            if mark:
                self.state.indent_marks.pop()

    @contextmanager
    def rich(
        self, style: str | None = None, highlight: bool = False
    ) -> Generator[PrintState, Any, None]:
        """set the rich style and highlight, and restore them after the context.

        Args:
            style(str | None): style to use, default to None
            highlight(bool): whether to highlight the text, default to False

        Yields:
            PrintState: the state with the new style and highlight.
        """
        old_style = self.state.rich_style
        old_highlight = self.state.rich_highlight
        self.state.rich_style = style
        self.state.rich_highlight = highlight
        try:
            yield self.state
        finally:
            self.state.rich_style = old_style
            self.state.rich_highlight = old_highlight

    @contextmanager
    def string_io(self) -> Generator[io.StringIO, Any, None]:
        """Temporary string IO for capturing output.

        Yields:
            io.StringIO: the string IO object.
        """
        stream = io.StringIO()
        old_file = self.console.file
        self.console.file = stream
        try:
            yield stream
        finally:
            self.console.file = old_file
            stream.close()

    def result_str(self, results: list["ir.ResultValue"]) -> str:
        """return the string representation of a list of result values.

        Args:
            results(list[ir.ResultValue]): list of result values to print
        """
        with self.string_io() as stream:
            self.print_seq(results, delim=", ")
            result_str = stream.getvalue()
        return result_str

    def debug(self, message: str):
        """Print a debug message."""
        self.state.messages.append(f"DEBUG: {message}")



from __future__ import annotations

from dataclasses import field, dataclass

from kirin import ir
from kirin.rewrite.abc import RewriteRule, RewriteResult

try:
    hash(slice(None))
    _HASHABLE_SLICE = True
except TypeError:
    _HASHABLE_SLICE = False


@dataclass
class Info:
    """An object to hold the comparison information of a statement."""

    head: type[ir.Statement]
    args: tuple[ir.SSAValue, ...]
    attributes: tuple[ir.Attribute, ...]
    successors: tuple[ir.Block, ...]
    regions: tuple[ir.Region, ...]
    _hash: int = field(init=False, repr=False)
    _hashable: bool = field(init=False, repr=False)

    def __post_init__(self):
        if not _HASHABLE_SLICE and any(
            isinstance(attr, ir.PyAttr) and isinstance(attr.data, slice)
            for attr in self.attributes
        ):
            self._hash = id(self)
            self._hashable = False
        else:
            self._hash = hash(
                (id(self.head),)
                + tuple(id(ssa) for ssa in self.args)
                + tuple(hash(attr) for attr in self.attributes)
                + tuple(id(succ) for succ in self.successors)
                + tuple(id(region) for region in self.regions)
            )
            self._hashable = True

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: object) -> bool:
        return self is other or (
            self._hashable
            and isinstance(other, Info)
            and self.head == other.head
            and self.args == other.args
            and self.attributes == other.attributes
            and self.successors == other.successors
            and self.regions == other.regions
        )


@dataclass
class CommonSubexpressionElimination(RewriteRule):

    def rewrite_Block(self, node: ir.Block) -> RewriteResult:
        seen: dict[Info, ir.Statement] = {}

        for stmt in node.stmts:
            if not stmt.has_trait(ir.Pure):
                continue

            if stmt.regions:
                continue

            info = Info(
                head=type(stmt),
                args=tuple(stmt.args),
                attributes=tuple(stmt.attributes.values()),
                successors=tuple(stmt.successors),
                regions=tuple(stmt.regions),
            )
            if info in seen:
                old_stmt = seen[info]
                for result, old_result in zip(stmt._results, old_stmt.results):
                    result.replace_by(old_result)
                stmt.delete()
                return RewriteResult(has_done_something=True)
            else:
                seen[info] = stmt
        return RewriteResult()

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not node.regions:
            return RewriteResult()

        has_done_something = False
        for region in node.regions:
            for block in region.blocks:
                result = self.rewrite_Block(block)
                if result.has_done_something:
                    has_done_something = True

        return RewriteResult(has_done_something=has_done_something)



from typing import Any

from kirin import ir, types
from kirin.decl import info, statement

T = types.TypeVar("T")


@statement(repr=True)
class DummyStatement(ir.Statement):
    name = "constant"
    traits = frozenset({ir.Pure(), ir.ConstantLike()})

    # args
    noinfo: ir.SSAValue
    vararg_noinfo: tuple[ir.SSAValue, ...]
    xxx: ir.SSAValue = info.argument(T)
    xxx_Any: ir.SSAValue = info.argument()
    xxx_vararg: tuple[ir.SSAValue, ...] = info.argument()

    # results
    xxx_result: ir.ResultValue = info.result(T)

    # attributes
    xxx_property: Any = info.attribute(T, default="")
    xxx_attribute: ir.PyAttr[float] = info.attribute()
    xxx_dict: dict[str, int] = info.attribute()

    # regions
    xxx_region_noinfo: ir.Region = info.region()
    xxx_region: ir.Region = info.region(default_factory=ir.Region)
    xxx_region_multi: ir.Region = info.region(default_factory=ir.Region, multi=True)

    # blocks
    block_noinfo: ir.Block = info.block()
    block_default: ir.Block = info.block(default_factory=ir.Block)


def test_init():
    args = [ir.TestValue() for _ in range(8)]
    stmt = DummyStatement(
        args[0],
        (args[1], args[2]),
        args[3],
        xxx_Any=args[4],
        xxx_vararg=(args[5], args[6]),
        xxx_attribute=ir.PyAttr(2),
        xxx_property=1,
        xxx_dict={"a": 1},
        xxx_region_noinfo=ir.Region(),
        block_noinfo=ir.Block(),
    )

    print(stmt)

    assert stmt.noinfo is args[0]
    assert stmt.vararg_noinfo[0] is args[1]
    assert stmt.vararg_noinfo[1] is args[2]
    assert stmt.xxx is args[3]
    assert stmt.xxx_Any is args[4]
    assert stmt.xxx_vararg[0] is args[5]
    assert stmt.xxx_vararg[1] is args[6]

    assert isinstance(stmt.xxx_attribute, ir.PyAttr)
    assert stmt.xxx_attribute.data == 2
    assert stmt.xxx_property == 1
    assert isinstance(stmt.xxx_dict, dict)
    assert stmt.xxx_dict == {"a": 1}

    assert isinstance(stmt.xxx_region_noinfo, ir.Region)
    assert isinstance(stmt.xxx_region, ir.Region)
    assert isinstance(stmt.xxx_region_multi, ir.Region)
    assert isinstance(stmt.block_noinfo, ir.Block)
    assert isinstance(stmt.block_default, ir.Block)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar
from dataclasses import dataclass

if TYPE_CHECKING:
    from kirin import ir, interp
    from kirin.graph import Graph


IRNodeType = TypeVar("IRNodeType")


@dataclass(frozen=True)
class Trait(ABC, Generic[IRNodeType]):
    """Base class for all statement traits."""

    def verify(self, node: IRNodeType):
        pass


@dataclass(frozen=True)
class AttrTrait(Trait["ir.Attribute"]):
    """Base class for all attribute traits."""

    def verify(self, node: ir.Attribute):
        pass


@dataclass(frozen=True)
class StmtTrait(Trait["ir.Statement"], ABC):
    """Base class for all statement traits."""

    def verify(self, node: ir.Statement):
        pass


GraphType = TypeVar("GraphType", bound="Graph[ir.Block]")


@dataclass(frozen=True)
class RegionGraph(StmtTrait, Generic[GraphType]):
    """A trait that indicates the properties of the statement's region."""

    @abstractmethod
    def get_graph(self, region: ir.Region) -> GraphType: ...


@dataclass(frozen=True)
class RegionInterpretationTrait(StmtTrait):
    """A trait that indicates the execution convention of the statement's region.

    For example, a region is SSA CFG if it follows SSA form and has a control flow graph.
    This trait also indicates that there is an available implementation of the trait in each
    interpreter.
    """

    ValueType = TypeVar("ValueType")

    @classmethod
    @abstractmethod
    def set_region_input(
        cls,
        frame: interp.FrameABC[ir.SSAValue, ValueType],
        region: ir.Region,
        *inputs: ValueType,
    ) -> None: ...



import typing
from collections.abc import Sequence

from kirin import ir, types
from kirin.dialects.func.attrs import Signature
from kirin.dialects.ilist.runtime import IList
from kirin.serialization.base.context import (
    MethodSymbolMeta,
    SerializationContext,
    mangle,
    get_str_from_type,
)
from kirin.serialization.base.serializable import Serializable
from kirin.serialization.base.serializationunit import SerializationUnit
from kirin.serialization.base.serializationmodule import SerializationModule


class Serializer:
    _ctx: SerializationContext

    def __init__(self) -> None:
        self._ctx = SerializationContext()
        self._ctx.clear()

    def encode(self, obj: ir.Method) -> SerializationModule:
        body = self.serialize_method(obj)
        if getattr(self._ctx, "Method_Symbol", None):
            st: dict[str, MethodSymbolMeta] = {}
            for mangled, meta in self._ctx.Method_Symbol.items():
                sym_name = meta.get("sym_name")
                if sym_name is None:
                    raise ValueError(f"symbol_table[{mangled}] missing 'sym_name'")
                st[mangled] = (
                    MethodSymbolMeta(
                        sym_name=sym_name,
                        arg_types=meta.get("arg_types", []),
                    )
                    if isinstance(meta, dict)
                    else meta
                )
            symbol_table: dict[str, MethodSymbolMeta] = st
        else:
            symbol_table = dict[str, MethodSymbolMeta]()
        return SerializationModule(symbol_table=symbol_table, body=body)

    def serialize(
        self, obj: Serializable | Sequence[typing.Any] | None
    ) -> SerializationUnit:
        ser_method = getattr(
            self, "serialize_" + type(obj).__name__.lower(), self.generic_serialize
        )
        return ser_method(obj)

    def generic_serialize(self, obj: object) -> SerializationUnit:
        if isinstance(obj, bool):
            return self.serialize_boolean(obj)
        elif isinstance(obj, bytes):
            return self.serialize_bytes(obj)
        elif isinstance(obj, bytearray):
            return self.serialize_bytes_array(obj)
        elif isinstance(obj, dict):
            return self.serialize_dict(obj)
        elif isinstance(obj, float):
            return self.serialize_float(obj)
        elif isinstance(obj, frozenset):
            return self.serialize_frozenset(obj)
        elif isinstance(obj, int):
            return self.serialize_int(obj)
        elif isinstance(obj, list):
            return self.serialize_list(obj)
        elif isinstance(obj, range):
            return self.serialize_range(obj)
        elif isinstance(obj, set):
            return self.serialize_set(obj)
        elif isinstance(obj, slice):
            return self.serialize_slice(obj)
        elif isinstance(obj, str):
            return self.serialize_str(obj)
        elif obj is None:
            return self.serialize_none(obj)
        elif isinstance(obj, tuple):
            return self.serialize_tuple(obj)
        elif isinstance(obj, type):
            return self.serialize_type(obj)
        else:
            raise ValueError(
                f"Unsupported object type {type(obj)} for serialization. Implement 'serialize' method."
            )

    def serialize_method(self, mthd: ir.Method) -> SerializationUnit:

        mangled = mangle(
            mthd.sym_name,
            getattr(mthd, "arg_types", ()),
            getattr(mthd, "ret_type", None),
        )
        arg_types_list: list[str] = []
        ret_type: str = get_str_from_type(mthd.return_type)
        for t in getattr(mthd, "arg_types", ()):
            arg_types_list.append(get_str_from_type(t))
        if mthd.sym_name is None:
            raise ValueError("Method.sym_name is None, cannot serialize.")
        meta: MethodSymbolMeta = {
            "sym_name": mthd.sym_name,
            "arg_types": arg_types_list,
            "ret_type": ret_type,
        }

        existing = self._ctx.Method_Symbol.get(mangled)
        if existing is not None:
            if existing != meta:
                raise ValueError(
                    f"Mangled name collision for {mangled}: existing={existing} new={meta}"
                )
        else:
            self._ctx.Method_Symbol[mangled] = meta

        out = SerializationUnit(
            kind="method",
            module_name=ir.Method.__module__,
            class_name=ir.Method.__name__,
            data={
                "sym_name": mthd.sym_name,
                "arg_names": mthd.arg_names,
                "dialects": self.serialize_dialect_group(mthd.dialects),
                "code": self.serialize_statement(mthd.code),
                "mangled": mangled,
            },
        )
        return out

    def serialize_statement(self, stmt: ir.Statement) -> SerializationUnit:
        return SerializationUnit(
            kind="statement",
            module_name=stmt.__class__.__module__,
            class_name=stmt.__class__.__name__,
            data={
                "id": self._ctx.stmt_idtable[stmt],
                "dialect": self.serialize(stmt.dialect),
                "name": self.serialize_str(stmt.name),
                "_args": self.serialize_tuple(stmt._args),
                "_results": self.serialize_list(stmt._results),
                "_name_args_slice": self.serialize_dict(stmt._name_args_slice),
                "attributes": self.serialize_dict(stmt.attributes),
                "successors": self.serialize_list(stmt.successors),
                "_regions": self.serialize_list(stmt._regions),
            },
        )

    def serialize_blockargument(self, arg: ir.BlockArgument) -> SerializationUnit:
        return SerializationUnit(
            kind="block-arg",
            module_name=ir.BlockArgument.__module__,
            class_name=ir.BlockArgument.__name__,
            data={
                "kind": "block-arg",
                "id": self._ctx.ssa_idtable[arg],
                "blk_id": self._ctx.blk_idtable[arg.owner],
                "index": arg.index,
                "type": self.serialize_attribute(arg.type),
                "name": arg.name,
            },
        )

    def serialize_region(self, region: ir.Region) -> SerializationUnit:
        region_id = self._ctx.region_idtable[region]
        if region_id in self._ctx.Region_Lookup:
            return SerializationUnit(
                kind="region_ref",
                module_name=ir.Region.__module__,
                class_name=ir.Region.__name__,
                data={
                    "kind": "region_ref",
                    "id": region_id,
                },
            )
        else:
            self._ctx.Region_Lookup[region_id] = region
            return SerializationUnit(
                kind="region",
                module_name=ir.Region.__module__,
                class_name=ir.Region.__name__,
                data={
                    "id": region_id,
                    "blocks": [self.serialize(block) for block in region.blocks],
                },
            )

    def serialize_block(self, block: ir.Block) -> SerializationUnit:
        if self._ctx.blk_idtable[block] in self._ctx.Block_Lookup:
            return SerializationUnit(
                kind="block_ref",
                module_name=ir.Block.__module__,
                class_name=ir.Block.__name__,
                data={
                    "id": self._ctx.blk_idtable[block],
                },
            )
        else:
            self._ctx.Block_Lookup[self._ctx.blk_idtable[block]] = block
            return SerializationUnit(
                kind="block",
                module_name=ir.Block.__module__,
                class_name=ir.Block.__name__,
                data={
                    "id": self._ctx.blk_idtable[block],
                    "stmts": [self.serialize_statement(stmt) for stmt in block.stmts],
                    "_args": [self.serialize_blockargument(arg) for arg in block.args],
                },
            )

    def serialize_boolean(self, value: bool) -> SerializationUnit:
        return SerializationUnit(
            kind="bool",
            module_name=bool.__module__,
            class_name=bool.__name__,
            data={
                "value": str(value) if value else "",
            },
        )

    def serialize_bytes(self, value: bytes) -> SerializationUnit:
        return SerializationUnit(
            kind="bytes",
            module_name=bytes.__module__,
            class_name=bytes.__name__,
            data={
                "value": value.hex(),
            },
        )

    def serialize_bytes_array(self, value: bytearray) -> SerializationUnit:
        return SerializationUnit(
            kind="bytearray",
            module_name=bytearray.__module__,
            class_name=bytearray.__name__,
            data={
                "value": bytes(value).hex(),
            },
        )

    def serialize_dict(self, value: dict) -> SerializationUnit:
        return SerializationUnit(
            kind="dict",
            module_name=dict.__module__,
            class_name=dict.__name__,
            data={
                "keys": [self.serialize(k) for k in value.keys()],
                "values": [self.serialize(v) for v in value.values()],
            },
        )

    def serialize_float(self, value: float) -> SerializationUnit:
        return SerializationUnit(
            kind="float",
            module_name=float.__module__,
            class_name=float.__name__,
            data={
                "value": str(value),
            },
        )

    def serialize_frozenset(self, value: frozenset) -> SerializationUnit:
        return SerializationUnit(
            kind="frozenset",
            module_name=frozenset.__module__,
            class_name=frozenset.__name__,
            data={
                "value": [self.serialize(x) for x in value],
            },
        )

    def serialize_int(self, value: int) -> SerializationUnit:
        return SerializationUnit(
            kind="int",
            module_name=int.__module__,
            class_name=int.__name__,
            data={
                "value": str(value),
            },
        )

    def serialize_list(self, value: list) -> SerializationUnit:
        return SerializationUnit(
            kind="list",
            module_name=list.__module__,
            class_name=list.__name__,
            data={
                "value": [self.serialize(x) for x in value],
            },
        )

    def serialize_range(self, r: range) -> SerializationUnit:
        return SerializationUnit(
            kind="range",
            module_name=range.__module__,
            class_name=range.__name__,
            data={
                "start": self.serialize_int(r.start),
                "stop": self.serialize_int(r.stop),
                "step": self.serialize_int(r.step),
            },
        )

    def serialize_set(self, value: set) -> SerializationUnit:
        return SerializationUnit(
            kind="set",
            module_name=set.__module__,
            class_name=set.__name__,
            data={
                "value": [self.serialize(x) for x in value],
            },
        )

    def serialize_slice(self, value: slice) -> SerializationUnit:
        return SerializationUnit(
            kind="slice",
            module_name=slice.__module__,
            class_name=slice.__name__,
            data={
                "start": self.serialize(value.start),
                "stop": self.serialize(value.stop),
                "step": self.serialize(value.step),
            },
        )

    def serialize_str(self, value: str) -> SerializationUnit:
        return SerializationUnit(
            kind="str",
            module_name=str.__module__,
            class_name=str.__name__,
            data={
                "value": value,
            },
        )

    def serialize_none(self, value: None) -> SerializationUnit:
        return SerializationUnit(
            kind="none",
            module_name=type(value).__module__,
            class_name=type(value).__name__,
            data={},
        )

    def serialize_tuple(self, value: tuple) -> SerializationUnit:
        return SerializationUnit(
            kind="tuple",
            module_name=tuple.__module__,
            class_name=tuple.__name__,
            data={
                "value": [self.serialize(x) for x in value],
            },
        )

    def serialize_attribute(self, attr: ir.Attribute) -> SerializationUnit:
        if not isinstance(attr, Serializable):
            raise TypeError(f"Attribute {attr} is not Serializable.")
        return attr.serialize(self)

    def serialize_resultvalue(self, result: ir.ResultValue) -> SerializationUnit:
        return SerializationUnit(
            kind="result-value",
            module_name=ir.ResultValue.__module__,
            class_name=ir.ResultValue.__name__,
            data={
                "kind": "result-value",
                "id": self._ctx.ssa_idtable[result],
                "owner": self.serialize_str(self._ctx.stmt_idtable[result.owner]),
                "index": result.index,
                "type": self.serialize_attribute(result.type),
                "name": result.name,
            },
        )

    def serialize_type(self, typ: type) -> SerializationUnit:
        return SerializationUnit(
            kind="type", module_name=typ.__module__, class_name=typ.__name__, data={}
        )

    def serialize_dialect(self, dialect: ir.Dialect) -> SerializationUnit:
        return SerializationUnit(
            kind="dialect",
            module_name=ir.Dialect.__module__,
            class_name=ir.Dialect.__name__,
            data={
                "name": self.serialize_str(dialect.name),
                "stmts": self.serialize_list(dialect.stmts),
            },
        )

    def serialize_dialect_group(self, group: ir.DialectGroup) -> SerializationUnit:
        return SerializationUnit(
            kind="dialect_group",
            module_name=ir.DialectGroup.__module__,
            class_name=ir.DialectGroup.__name__,
            data={
                "data": self.serialize_frozenset(group.data),
            },
        )

    def serialize_pyclass(self, pc: types.PyClass) -> SerializationUnit:
        return SerializationUnit(
            kind="pyclass",
            module_name=pc.__module__,
            class_name=pc.__class__.__name__,
            data=dict(
                typ=self.serialize_type(pc.typ),
                display_name=self.serialize_str(pc.display_name),
                prefix=self.serialize_str(pc.prefix),
            ),
        )

    def serialize_typevar(self, tv: types.TypeVar) -> SerializationUnit:
        return SerializationUnit(
            kind="typevar",
            module_name=tv.__module__,
            class_name=tv.__class__.__name__,
            data={
                "varname": self.serialize_str(tv.varname),
                "bound": self.serialize_attribute(tv.bound),
            },
        )

    def serialize_generic(self, gen: types.Generic) -> SerializationUnit:
        return SerializationUnit(
            kind="generic",
            module_name=gen.__module__,
            class_name=gen.__class__.__name__,
            data={
                "body": self.serialize_pyclass(gen.body),
                "vars": self.serialize_tuple(gen.vars),
                "vararg": self.serialize(gen.vararg),
            },
        )

    def serialize_vararg(self, var: types.Vararg) -> SerializationUnit:
        return SerializationUnit(
            kind="vararg",
            module_name=var.__module__,
            class_name=var.__class__.__name__,
            data={"typ": self.serialize_attribute(var.typ)},
        )

    def serialize_anytype(self, at: types.AnyType) -> SerializationUnit:
        return SerializationUnit(
            kind="anytype",
            module_name=at.__module__,
            class_name=at.__class__.__name__,
            data=dict(),
        )

    def serialize_bottomtype(self, bt: types.BottomType) -> SerializationUnit:
        return SerializationUnit(
            kind="bottomtype",
            module_name=bt.__module__,
            class_name=bt.__class__.__name__,
            data=dict(),
        )

    def serialize_pyattr(self, pa: ir.PyAttr) -> SerializationUnit:
        return SerializationUnit(
            kind="pyattr",
            module_name=pa.__module__,
            class_name=pa.__class__.__name__,
            data={
                "data": self.serialize(pa.data),
                "pytype": self.serialize_attribute(pa.type),
            },
        )

    def serialize_literal(self, lit: types.Literal) -> SerializationUnit:
        return SerializationUnit(
            kind="literal",
            module_name=lit.__module__,
            class_name=lit.__class__.__name__,
            data={
                "value": self.serialize(lit.data),
                "type": self.serialize_attribute(lit.type),
            },
        )

    def serialize_union(self, uni: types.Union) -> SerializationUnit:
        return SerializationUnit(
            kind="Union",
            module_name=uni.__module__,
            class_name=uni.__class__.__name__,
            data={"types": self.serialize_frozenset(uni.types)},
        )

    def serialize_signature(self, sig: Signature) -> SerializationUnit:
        return SerializationUnit(
            kind="signature",
            module_name=sig.__module__,
            class_name=sig.__class__.__name__,
            data={
                "inputs": self.serialize_tuple(sig.inputs),
                "output": self.serialize(sig.output),
            },
        )

    def serialize_ilist(self, ilist: IList) -> SerializationUnit:
        return SerializationUnit(
            kind="ilist",
            module_name=ilist.__module__,
            class_name=ilist.__class__.__name__,
            data={
                "data": self.serialize(ilist.data),
                "elem": self.serialize_attribute(ilist.elem),
            },
        )



from __future__ import annotations

import sys
import inspect
import textwrap
from typing import TYPE_CHECKING

from rich.console import Console

from kirin.exception import StaticCheckError
from kirin.print.printer import Printer

if TYPE_CHECKING:
    from kirin.ir import IRNode, Method


class ValidationError(StaticCheckError):
    def __init__(self, node: "IRNode", *messages: str, help: str | None = None) -> None:
        super().__init__(*messages, help=help)
        self.node = node
        self.source = node.source
        self.method: Method | None = None

    def attach(self, method: Method):
        if self.method:
            return
        self.method = method

        from kirin.ir.nodes.stmt import Statement

        console = Console(force_terminal=True, force_jupyter=False, file=sys.stderr)
        printer = Printer(console=console)
        # NOTE: populate the printer with the method body
        with printer.string_io():
            printer.print(method.code)
        with printer.string_io() as io:
            printer.print(self.node)
            node_str = io.getvalue()

        node_str = "\n".join(
            map(lambda each_line: " " * 4 + each_line, node_str.splitlines())
        )
        if isinstance(self.node, Statement):
            dialect = self.node.dialect.name if self.node.dialect else "<no dialect>"
            self.args += (
                "when verifying the following statement",
                f" `{dialect}.{type(self.node).__name__}` at\n",
                f"{node_str}\n",
            )
        else:
            self.args += (
                f"when verifying the following statement `{type(self.node).__name__}` at\n",
                f"{node_str}\n",
            )

        if self.source:
            self.source.lineno_begin = method.lineno_begin

        if self.node.source and method.py_func:  # print hint if we have a source
            source = textwrap.dedent(inspect.getsource(method.py_func))
            self.lines = source.splitlines()
            if self.source and self.source.file is None:
                self.source.file = method.file


class TypeCheckError(ValidationError):
    pass


class CompilerError(Exception):
    pass



import json
from typing import Any

from kirin.serialization.base.serializationunit import SerializationUnit
from kirin.serialization.base.serializationmodule import SerializationModule


class JSONSerializer:
    def _to_jsonifiable(self, obj: Any) -> Any:
        if isinstance(obj, SerializationModule):
            return {
                "__serialization_module__": True,
                "symbol_table": self._to_jsonifiable(obj.symbol_table),
                "body": self._to_jsonifiable(obj.body),
            }
        if isinstance(obj, SerializationUnit):
            return {
                "__serialization_unit__": True,
                "kind": obj.kind,
                "module_name": obj.module_name,
                "class_name": obj.class_name,
                "data": self._to_jsonifiable(obj.data),
            }
        if isinstance(obj, dict):
            return {k: self._to_jsonifiable(v) for k, v in obj.items()}
        if isinstance(obj, (list, tuple)):
            return [self._to_jsonifiable(v) for v in obj]
        return obj

    def _from_jsonifiable(self, obj: Any) -> Any:
        if isinstance(obj, dict):
            if obj.get("__serialization_module__"):
                symbol_table = self._from_jsonifiable(obj.get("symbol_table", {}))
                body = self._from_jsonifiable(obj.get("body"))
                return SerializationModule(symbol_table=symbol_table, body=body)
            if obj.get("__serialization_unit__"):
                data = self._from_jsonifiable(obj.get("data", {}))
                return SerializationUnit(
                    kind=obj["kind"],
                    module_name=obj["module_name"],
                    class_name=obj["class_name"],
                    data=data,
                )
            return {k: self._from_jsonifiable(v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [self._from_jsonifiable(v) for v in obj]
        return obj

    def encode(self, data: SerializationModule) -> str:
        payload = self._to_jsonifiable(data)
        return json.dumps(payload, separators=(",", ":"), ensure_ascii=False)

    def decode(self, data: str) -> SerializationModule:
        parsed = json.loads(data)
        result = self._from_jsonifiable(parsed)
        if not isinstance(result, SerializationModule):
            raise TypeError("decoded payload is not a SerializationModule")
        return result

    def serialize_serialization_unit(self, unit: SerializationUnit) -> str:
        payload = self._to_jsonifiable(unit)
        return json.dumps(payload, separators=(",", ":"), ensure_ascii=False)

    def deserialize_serialization_unit(self, data: str) -> SerializationUnit:
        parsed = json.loads(data)
        obj = self._from_jsonifiable(parsed)
        if not isinstance(obj, SerializationUnit):
            raise TypeError("decoded payload is not a SerializationUnit")
        return obj



"""Constant statement for Python dialect.

This module contains the dialect for the Python `constant` statement, including:

- The `Constant` statement class.
- The lowering pass for the `constant` statement.
- The concrete implementation of the `constant` statement.
- The Julia emitter for the `constant` statement.

This dialect maps `ast.Constant` nodes to the `Constant` statement.
"""

from __future__ import annotations

import ast
from typing import Generic, TypeVar

from kirin import ir, emit, types, interp, lowering
from kirin.decl import info, statement
from kirin.print import Printer

dialect = ir.Dialect("py.constant")

T = TypeVar("T", covariant=True)


@statement(dialect=dialect)
class Constant(ir.Statement, Generic[T]):
    name = "constant"
    traits = frozenset({ir.Pure(), ir.ConstantLike(), lowering.FromPythonCall()})
    value: ir.Data[T] = info.attribute()
    result: ir.ResultValue = info.result()

    # NOTE: we allow py.Constant take data.PyAttr too
    def __init__(self, value: T | ir.Data[T]) -> None:
        if not isinstance(value, ir.Data):
            value = ir.PyAttr(value)
        super().__init__(
            attributes={"value": value},
            result_types=(value.type,),
        )

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        if isinstance(self.value, ir.PyAttr):
            printer.plain_print(repr(self.value.data))
        else:  # other attributes
            printer.plain_print(repr(self.value))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.result.type)

    def check_type(self) -> None:
        if not isinstance(self.result.type, types.TypeAttribute):
            raise TypeError(
                f"Expected result type to be PyType, got {self.result.type}"
            )


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Constant(
        self, state: lowering.State, node: ast.Constant
    ) -> lowering.Result:
        return state.current_frame.push(
            Constant(node.value),
        )


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Constant)
    def constant(self, interp, frame: interp.Frame, stmt: Constant):
        return (stmt.value.unwrap(),)


@dialect.register(key="emit.julia")
class Julia(interp.MethodTable):

    @interp.impl(Constant)
    def constant(self, emit: emit.Julia, frame: emit.JuliaFrame, stmt: Constant):
        return (emit.get_attribute(frame, stmt.value),)



from abc import abstractmethod
from typing import TYPE_CHECKING, Protocol, runtime_checkable

if TYPE_CHECKING:
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.serializationunit import SerializationUnit


@runtime_checkable
class Serializable(Protocol):
    @abstractmethod
    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        raise NotImplementedError



from kirin import ir, types
from kirin.analysis import const
from kirin.dialects import py, ilist
from kirin.rewrite.abc import RewriteRule, RewriteResult


class FlattenAdd(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not (
            isinstance(node, py.binop.Add)
            and node.lhs.type.is_subseteq(ilist.IListType)
            and node.rhs.type.is_subseteq(ilist.IListType)
        ):
            return RewriteResult()

        # check if we are adding two ilist.New objects
        new_data = ()

        # lhs:
        if isinstance((lhs := node.lhs).owner, ilist.New):
            new_data += lhs.owner.values
        elif (
            not isinstance(const_lhs := lhs.hints.get("const"), const.Value)
            or len(const_lhs.data) > 0
        ):
            return RewriteResult()

        # rhs:
        if isinstance((rhs := node.rhs).owner, ilist.New):
            new_data += rhs.owner.values
        elif (
            not isinstance(const_rhs := rhs.hints.get("const"), const.Value)
            or len(const_rhs.data) > 0
        ):
            return RewriteResult()

        assert isinstance(rhs_type := rhs.type, types.Generic), "Impossible"
        assert isinstance(lhs_type := lhs.type, types.Generic), "Impossible"

        lhs_elem_type = lhs_type.vars[0]
        rhs_elem_type = rhs_type.vars[0]

        result_elem_type = lhs_elem_type.join(rhs_elem_type)
        node.replace_by(ilist.New(values=new_data, elem_type=result_elem_type))

        return RewriteResult(has_done_something=True)



from typing import cast

from kirin import types
from kirin.prelude import basic
from kirin.analysis import const
from kirin.dialects.py.range import Range


@basic
def new_range(a: int, b: int, c: int):
    x = range(a)
    y = range(a, b)
    z = range(a, b, c)
    return x, y, z


new_range.print()


def test_new_range():
    stmt = cast(Range, new_range.callable_region.blocks[0].stmts.at(2))
    assert isinstance(hint := stmt.start.hints.get("const"), const.Value)
    assert hint.data == 0
    assert stmt.stop.type.is_subseteq(types.Int)
    assert isinstance(hint := stmt.step.hints.get("const"), const.Value)
    assert hint.data == 1

    stmt = cast(Range, new_range.callable_region.blocks[0].stmts.at(4))
    assert stmt.start.type.is_subseteq(types.Int)
    assert stmt.stop.type.is_subseteq(types.Int)
    assert stmt.step.type.is_subseteq(types.Int)
    assert isinstance(hint := stmt.step.hints.get("const"), const.Value)
    assert hint.data == 1

    stmt = cast(Range, new_range.callable_region.blocks[0].stmts.at(5))
    assert stmt.start.type.is_subseteq(types.Int)
    assert stmt.stop.type.is_subseteq(types.Int)
    assert stmt.step.type.is_subseteq(types.Int)



from __future__ import annotations

import inspect
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Union,
    Generic,
    TypeVar,
    Callable,
    ParamSpec,
    Concatenate,
    cast,
    overload,
)
from functools import update_wrapper
from dataclasses import field, dataclass
from collections.abc import Iterable

from typing_extensions import Self

from kirin.ir.method import Method
from kirin.ir.traits import SymbolTable, SymbolOpInterface
from kirin.ir.exception import CompilerError, ValidationError

if TYPE_CHECKING:
    from kirin.ir import Dialect, Statement
    from kirin.lowering import Python
    from kirin.registry import Registry
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer
    from kirin.serialization.base.serializationunit import SerializationUnit

PassParams = ParamSpec("PassParams")
RunPass = Callable[Concatenate[Method, PassParams], None]
RunPassGen = Callable[["DialectGroup"], RunPass[PassParams]]


@dataclass(init=False)
class DialectGroup(Generic[PassParams]):
    # method wrapper params
    Param = ParamSpec("Param")
    RetType = TypeVar("RetType")
    MethodTransform = Callable[[Callable[Param, RetType]], Method[Param, RetType]]

    data: frozenset["Dialect"]
    """The set of dialects in the group."""
    # NOTE: this is used to create new dialect groups from existing one
    run_pass_gen: RunPassGen[PassParams] | None = None
    """the function that generates the `run_pass` function.

    This is used to create new dialect groups from existing ones, while
    keeping the same `run_pass` function.
    """
    run_pass: RunPass[PassParams] | None = None
    """the function that runs the passes on the method."""

    lowering: Python = field(init=False, repr=False)
    """the lowering object used to lower the method."""

    symbol_table: dict[str, Statement] = field(
        default_factory=dict, init=False, repr=False
    )

    def __init__(
        self,
        dialects: Iterable[Union["Dialect", ModuleType]],
        run_pass: RunPassGen[PassParams] | None = None,
    ):
        self.symbol_table = {}
        self.data = frozenset(self.map_module(dialect) for dialect in dialects)
        if run_pass is None:
            self.run_pass_gen = None
            self.run_pass = None
        else:
            self.run_pass_gen = run_pass
            self.run_pass = run_pass(self)

        from kirin.lowering import Python

        self.lowering = Python(self)

    def __iter__(self):
        return iter(self.data)

    def __repr__(self) -> str:
        names = ", ".join(each.name for each in self.data)
        return f"DialectGroup([{names}])"

    @staticmethod
    def map_module(dialect: Union["Dialect", ModuleType]) -> "Dialect":
        """map the module to the dialect if it is a module.
        It assumes that the module has a `dialect` attribute
        that is an instance of [`Dialect`][kirin.ir.Dialect].
        """
        if isinstance(dialect, ModuleType):
            return getattr(dialect, "dialect")
        return dialect

    def add(self, dialect: Union["Dialect", ModuleType]) -> "DialectGroup":
        """add a dialect to the group.

        Args:
            dialect (Union[Dialect, ModuleType]): the dialect to add

        Returns:
            DialectGroup: the new dialect group with the added
        """
        return self.union([dialect])

    def union(self, dialect: Iterable[Union["Dialect", ModuleType]]) -> "DialectGroup":
        """union a set of dialects to the group.

        Args:
            dialect (Iterable[Union[Dialect, ModuleType]]): the dialects to union

        Returns:
            DialectGroup: the new dialect group with the union.
        """
        return DialectGroup(
            dialects=self.data.union(frozenset(self.map_module(d) for d in dialect)),
            run_pass=self.run_pass_gen,  # pass the run_pass_gen function
        )

    def discard(self, dialect: Union["Dialect", ModuleType]) -> "DialectGroup":
        """discard a dialect from the group.

        !!! note
            This does not raise an error if the dialect is not in the group.

        Args:
            dialect (Union[Dialect, ModuleType]): the dialect to discard

        Returns:
            DialectGroup: the new dialect group with the discarded dialect.
        """
        dialect_ = self.map_module(dialect)
        return DialectGroup(
            dialects=frozenset(
                each for each in self.data if each.name != dialect_.name
            ),
            run_pass=self.run_pass_gen,  # pass the run_pass_gen function
        )

    def __contains__(self, dialect) -> bool:
        """check if the dialect is in the group.

        Args:
            dialect (Union[Dialect, ModuleType]): the dialect to check.

        Returns:
            bool: True if the dialect is in the group, False otherwise.
        """
        return self.map_module(dialect) in self.data

    @property
    def registry(self) -> "Registry":
        """return the registry for the dialect group. This
        returns a proxy object that can be used to select
        the lowering interpreters, interpreters, and codegen
        for the dialects in the group.

        Returns:
            Registry: the registry object.
        """
        from kirin.registry import Registry

        return Registry(self)

    @overload
    def __call__(
        self,
        py_func: Callable[Param, RetType],
        *args: PassParams.args,
        **options: PassParams.kwargs,
    ) -> Method[Param, RetType]: ...

    @overload
    def __call__(
        self,
        py_func: None = None,
        *args: PassParams.args,
        **options: PassParams.kwargs,
    ) -> MethodTransform[Param, RetType]: ...

    def __call__(
        self,
        py_func: Callable[Param, RetType] | None = None,
        *args: PassParams.args,
        **options: PassParams.kwargs,
    ) -> Method[Param, RetType] | MethodTransform[Param, RetType]:
        """create a method from the python function.

        Args:
            py_func (Callable): the python function to create the method from.
            args (PassParams.args): the arguments to pass to the run_pass function.
            options (PassParams.kwargs): the keyword arguments to pass to the run_pass function.

        Returns:
            Method: the method created from the python function.
        """
        frame = inspect.currentframe()

        def wrapper(py_func: Callable) -> Method:
            if py_func.__name__ == "<lambda>":
                raise ValueError("Cannot compile lambda functions")

            lineno_offset, file = 0, ""
            mt = None
            if frame and frame.f_back is not None:
                call_site_frame = frame.f_back
                if py_func.__name__ in call_site_frame.f_locals:
                    mt = call_site_frame.f_locals[py_func.__name__]
                    if not isinstance(mt, Method):
                        raise CompilerError(
                            f"`{py_func.__name__}` is already defined in the current scope and is not a Method."
                        )

                lineno_offset = call_site_frame.f_lineno - 1
                file = call_site_frame.f_code.co_filename

            code = self.lowering.python_function(py_func, lineno_offset=lineno_offset)
            arg_names = ["#self#"] + inspect.getfullargspec(py_func).args

            if mt:
                mt.mod = inspect.getmodule(py_func)
                mt.dialects = self
                mt.code = code
                mt.py_func = py_func
                mt.nargs = len(arg_names)
                mt.arg_names = arg_names
                mt.sym_name = py_func.__name__
                mt.file = file
                mt.lineno_begin = lineno_offset
                mt.run_passes = self.run_pass
                mt.update_backedges()  # update the callee
                self.recompile_callers(mt)
            else:
                mt = Method(
                    dialects=self,
                    code=code,
                    nargs=len(arg_names),
                    mod=inspect.getmodule(py_func),
                    py_func=py_func,
                    sym_name=py_func.__name__,
                    arg_names=arg_names,
                    file=file,
                    lineno_begin=lineno_offset,
                )

            if doc := inspect.getdoc(py_func):
                mt.__doc__ = doc

            def run_pass(mt: Method) -> None:
                if self.run_pass is not None:
                    try:
                        self.run_pass(mt, *args, **options)
                    except ValidationError as e:
                        e.attach(mt)
                        raise e

            mt.run_passes = run_pass
            run_pass(mt)
            self.update_symbol_table(mt)
            return mt

        if py_func is not None:
            return wrapper(py_func)
        return wrapper

    def recompile_callers(self, method: Method) -> None:
        for caller in method.backedges:
            if caller.run_passes:
                caller.run_passes(caller)
            # propagate the changes to all callers
            caller.dialects.recompile_callers(caller)
        return

    def update_symbol_table(self, method: Method) -> None:
        trait = method.code.get_trait(SymbolTable)
        if trait is None:
            return

        for stmt in method.code.walk():
            trait = stmt.get_trait(SymbolOpInterface)
            if trait is None:
                continue

            name = trait.get_sym_name(stmt).unwrap()
            if name in self.symbol_table:
                raise ValidationError(
                    stmt,
                    f"duplicate symbol `{name}` in dialect group",
                )
            self.symbol_table[name] = stmt

    def is_structurally_equal(
        self, other: DialectGroup, context: dict | None = None
    ) -> bool:
        if not isinstance(other, DialectGroup):
            return False
        if len(self.data) != len(other.data):
            return False
        x = sorted(self.data, key=lambda d: d.name)
        y = sorted(other.data, key=lambda d: d.name)
        for dialect1, dialect2 in zip(x, y):
            if dialect1.name != dialect2.name:
                return False
        return True

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_dialect_group(self)

    @classmethod
    def deserialize(
        cls: type[Self], serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> Self:
        return cast(Self, deserializer.deserialize_dialect_group(serUnit))


def dialect_group(
    dialects: Iterable[Union["Dialect", ModuleType]],
) -> Callable[[RunPassGen[PassParams]], DialectGroup[PassParams]]:
    """Create a dialect group from the given dialects based on the
    definition of `run_pass` function.

    Args:
        dialects (Iterable[Union[Dialect, ModuleType]]): the dialects to include in the group.

    Returns:
        Callable[[RunPassGen[PassParams]], DialectGroup[PassParams]]: the dialect group.

    Example:
        ```python
        from kirin.dialects import cf, fcf, func, math

        @dialect_group([cf, fcf, func, math])
        def basic_no_opt(self):
            # initializations
            def run_pass(mt: Method) -> None:
                # how passes are applied to the method
                pass

            return run_pass
        ```
    """

    # NOTE: do not alias the annotation below
    def wrapper(
        transform: RunPassGen[PassParams],
    ) -> DialectGroup[PassParams]:
        ret = DialectGroup(dialects, run_pass=transform)
        update_wrapper(ret, transform)
        return ret

    return wrapper



"""The tuple dialect for Python.

This dialect provides a way to work with Python tuples in the IR, including:

- The `New` statement class.
- The lowering pass for the tuple statement.
- The concrete implementation of the tuple statement.
- The type inference implementation of the tuple addition with `py.binop.Add`.
- The constant propagation implementation of the tuple statement.
- The Julia emitter for the tuple statement.

This dialect maps `ast.Tuple` nodes to the `New` statement.
"""

import ast

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.analysis import const
from kirin.dialects.eltype import ElType
from kirin.dialects.py.binop import Add

dialect = ir.Dialect("py.tuple")


@statement(dialect=dialect)
class New(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    result: ir.ResultValue = info.result()

    def __init__(self, values: tuple[ir.SSAValue, ...]) -> None:
        result_type = types.Generic(tuple, *tuple(value.type for value in values))
        super().__init__(
            args=values,
            result_types=[
                result_type,
            ],
        )


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(New)
    def new(self, interp: interp.Interpreter, frame: interp.Frame, stmt: New):
        return (frame.get_values(stmt.args),)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(New)
    def new_tuple(
        self,
        interp,
        frame: interp.Frame[types.TypeAttribute],
        stmt: New,
    ):
        arg_types = frame.get_values(stmt.args)
        # arg_types should already be kirin compatible
        return (types.Generic(tuple, *arg_types),)

    @interp.impl(ElType, types.PyClass(tuple))
    def eltype_tuple(self, interp, frame: interp.Frame, stmt: ElType):
        tuple_type = frame.get(stmt.container)
        if isinstance(tuple_type, types.Generic):
            ret = tuple_type.vars[0]
            for var in tuple_type.vars[1:]:
                ret = ret.join(var)
            return (ret,)
        else:
            return (types.Any,)

    @interp.impl(Add, types.PyClass(tuple), types.PyClass(tuple))
    def add(self, interp, frame: interp.Frame[types.TypeAttribute], stmt):
        lhs = frame.get(stmt.lhs)
        rhs = frame.get(stmt.rhs)
        if isinstance(lhs, types.Generic) and isinstance(rhs, types.Generic):
            return (types.Generic(tuple, *(lhs.vars + rhs.vars)),)
        else:
            return (types.PyClass(tuple),)  # no type param, so unknown


@dialect.register(key="constprop")
class ConstPropTable(interp.MethodTable):

    @interp.impl(New)
    def new_tuple(
        self,
        _: const.Propagate,
        frame: const.Frame,
        stmt: New,
    ) -> interp.StatementResult[const.Result]:
        return (const.PartialTuple(tuple(x for x in frame.get_values(stmt.args))),)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Tuple(self, state: lowering.State, node: ast.Tuple) -> lowering.Result:
        return state.current_frame.push(
            New(tuple(state.lower(elem).expect_one() for elem in node.elts))
        )



from kirin.lattice.abc import (
    Lattice as Lattice,
    UnionMeta as UnionMeta,
    LatticeMeta as LatticeMeta,
    SingletonMeta as SingletonMeta,
    BoundedLattice as BoundedLattice,
)
from kirin.lattice.empty import EmptyLattice as EmptyLattice
from kirin.lattice.mixin import (
    IsSubsetEqMixin as IsSubsetEqMixin,
    SimpleJoinMixin as SimpleJoinMixin,
    SimpleMeetMixin as SimpleMeetMixin,
)
